# -*- coding: utf-8 -*-
import pandas as pd
import re
import ipywidgets as widgets
from IPython.display import clear_output


from citableclass.base import Citable
from .base import CorpusTextSearch
# from corpussearch.base import CorpusTextSearch


class SearchWordGUI(widgets.HBox):
    def __init__(self, optionList, colName):
        super(SearchWordGUI, self).__init__()
        self.optList = optionList
        self.column = colName
        if self.column not in self.optList:
            self.optList.extend([colName])

        #############
        # Functions #
        #############
        #def chgTxt(change):
        #    self._text.value = change['new']

        ###########
        # Widgets #
        ###########
        self._text = widgets.Text(description='Search for:', placeholder='term')
        self._text.layout.margin = '0 0 0 20px'

        self._select = widgets.Dropdown(
            description='Search in:',
            options=self.optList,
            value=self.column,
        )
        ###########
        # Actions #
        ###########
        #self._select.observe(
        #    chgTxt,
        #    names='value'
        #)

        children = [self._text, self._select]
        self.children = children

    @property
    def value(self):
        return self._text.value

    @property
    def part(self):
        return self._select.value


class CorpusGUI(CorpusTextSearch):

    def __init__(
            self, pathDF,
            dataType='pickle', dataIndex='multi', colname='text',
            maxValues=2500, pathMeta=False, pathType=False
            ):

        super(CorpusGUI, self).__init__(pathDF, dataType, dataIndex, colname, maxValues, pathMeta, pathType)
        self.initSearch = SearchWordGUI(self.colValueDictTrigger, self.column)
        self.accordion = widgets.Accordion(children=[self.initSearch])
        self.accordion.set_title(0, 'Enter search term')

        self.extendSearch = widgets.ToggleButtons(options=['more', 'less'])
        self.extendSearch.on_msg(self._addSearchField)

        self.searchButton = widgets.Button(
            description='Search',
            button_style='danger'
        )

        self.searchButton.on_msg(self._searchLogic)

        self.direction = widgets.ToggleButtons(options=['previous', 'next'])

        self.direction.on_msg(self._setSentence)

        self.outInfo = widgets.Output()

        self.outSentence = widgets.Textarea(
            value='',
            layout=widgets.Layout(flex='0 1 auto', height='200px', min_height='40px', width='70%'),
            placeholder='Sentence'
        )

        self.outMeta = widgets.Textarea(
            placeholder='Result',
            layout=widgets.Layout(flex='0 1 auto', height='200px', min_height='40px', width='30%'),
            #description='Result:',
            value=''
        )

    def _setDescription(self):
        res = 'Result {0}\n'.format(self.counter) + '\n'.join(
             ["{0}:\n {1}".format(x, y) for x, y in self.displayResult.iloc[self.counter].to_dict().items() if x != self.column]
        )
        return res

    def _setSentence(self, widget, content, buffers):
        if widget.value == 'previous':
            self.counter = self.counter - 1
        elif widget.value == 'next':
            self.counter = self.counter + 1
        if self.counter < self.displayResult.shape[0] and self.counter > -1:
            self.outSentence.value = self.displayResult[self.column].iloc[self.counter]
            self.outMeta.value = self._setDescription()
        else:
            self.outSentence.value = 'End of found results. Enter new search.'
        return

    def _addSearchField(self, widget, content, buffers):
        child = SearchWordGUI(self.colValueDictTrigger, self.column)
        children = []
        for ch in self.accordion.children:
            children.append(ch)
        if self.extendSearch.value == 'more':
            children.append(
                widgets.Dropdown(
                    description='Select logic to connect words',
                    options={'and': '&', 'or': '|', 'not': '~&'},
                    value='&'
                )
            )
            children.append(child)
            self.accordion.children = children
            self.accordion.set_title(len(children)-2, 'and/or/not')
            self.accordion.set_title(len(children)-1, 'add term')
        elif self.extendSearch.value == 'less':
            if len(children) >= 3:
                lesschildren = children[:-2]
                self.accordion.children = lesschildren
            else:
                pass
            pass
        return

    def _searchLogic(self, widget, content, buffers):
        self.tempRes = []
        self.tempList = []
        self.counter = 0
        for ch in self.accordion.children:
            if ch.value not in ['&', '|', '~&']:
                self.tempList.append((ch.part, ch.value))
            else:
                self.tempList.append(ch.value)
        self.outputResult = self.logicReduce(self.tempList)

        self.displayResult = self.outputResult.results()

        if self.displayResult.shape[0] > 0:
            # TODO: Add plotting logic in handle_submit(sender)
            # handle_submit(sender)

            self.outSentence.value = re.sub('\n|\s+', ' ', self.displayResult[self.column].iloc[self.counter])
            self.outMeta.value = self._setDescription()
            with self.outInfo:
                clear_output()
                print('Found {} entries.'.format(self.displayResult.shape[0]))
        else:
            with self.outInfo:
                clear_output()
                print('Found no entries. Try changing search!')

    def displayGUI(self):
        searchControl = widgets.HBox([self.extendSearch, self.searchButton, self.outInfo])
        textControl = widgets.HBox([self.direction])
        sentenceFields = widgets.HBox([self.outMeta, self.outSentence])
        searchBox = widgets.VBox([self.accordion, searchControl, textControl, sentenceFields])

        return display(searchBox)



# -*- coding: utf-8 -*-
import pandas as pd
import re
import difflib
import logging
import os
import tempfile
import gensim
import requests
from citableclass.base import Citable
from .base import CorpusTextSearch

logging.basicConfig(
    format='%(asctime)s : %(levelname)s : %(message)s',
    level=logging.INFO
    )


class CorpusML(CorpusTextSearch):

    def __init__(
            self, pathDF, language='english',
            dataType='pickle', dataIndex='multi', colname='text',
            maxValues=2500, pathMeta=False, pathType=False
            ):

        super(CorpusML, self).__init__(
            pathDF, dataType, dataIndex, colname,
            maxValues, pathMeta, pathType
            )

        self.model = gensim.models.Word2Vec(
            workers=4, min_count=5, size=300
            )

        self.model.random.seed(42)

        self.language = language

        if self.language == 'latin':
            from cltk.stem.lemma import LemmaReplacer
            from cltk.stem.latin.j_v import JVReplacer
            from cltk.stop.latin.stops import STOPS_LIST as stopwords
            from cltk.tokenize.word import nltk_tokenize_words as tokenizer
            self.jvreplacer = JVReplacer()
            lemmatizer = LemmaReplacer('latin')

        self.lemmatizer = lemmatizer
        self.tokenizer = tokenizer
        self.stopwords = stopwords

    def convert(self, row):
        reg = re.compile('[^a-zA-Z0-9]')
        if type(row) == str:
            sentence = self.tokenizer(row)
            sentence = [x.lower() for x in sentence]
            sentence = [x for x in sentence if x not in self.stopwords]
            sentence = [x for x in sentence if x and len(x) > 1]
            sentence = [reg.sub('', x) for x in sentence]
            sentence = [x.strip('0123456789') for x in sentence]
            if self.language == 'latin':
                sentence = [self.jvreplacer.replace(word) for word in sentence]
            try:
                sentence = self.lemmatizer.lemmatize(sentence)
            except:
                pass
            return sentence

    def saveTrainData(self):
        self.dataframe['training_data'] = self.dataframe[self.column].apply(
            lambda row: self.convert(row)
        )
        self.tempFile = tempfile.NamedTemporaryFile(delete=False)
        self.dataframe.to_pickle(self.tempFile.name)
        return

    def buildVocab(self):
        loadedData = pd.read_pickle(self.tempFile)
        self.training_data = loadedData.training_data.values.tolist()
        self.model.build_vocab(training_data)
        return

    def trainModel(self):
        self.model.train(
            self.training_data,
            total_examples=self.model.corpus_count,
            epochs=self.model.epochs
            )
        return

    def getSimilarContext(word):
        vocabulary = self.model.wv.vocab.keys()
        if word in vocabulary:
            res = self.model.wv.most_similar(word)
            return res
        else:
            simWord = difflib.get_close_matches(word, vocabulary, 1)
            print('Using similar word: {0}'.format(simWord[0]))
            res = self.model.wv.most_similar(simWord)
            return res



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import re
import difflib
from citableclass.base import Citable


class CorpusTextSearch(object):
    """
    This class is initialized with a searchstring, the name of the column,
    which contains the textstrings, and a path to a pickled dataframe.
    Optionally, a path to an excel file containing metadata can be provided.

    For other import formats change read_pickle and read_excel accordingly.

    Example:
            >>> search1 = SearchPhrases(
                            '[Ss]ol','text,'/path/to/kepler/dataframe'
                            )
            True
    """

    def __init__(
            self, pathDF,
            dataType='pickle', dataIndex='multi', colname='text',
            maxValues=2500, pathMeta=False, pathType=False
            ):

        self.datatype = dataType

        if pathType == 'DOI':
            pathDF = "http://dx.doi.org/{0}".format(pathDF)
        elif pathType == 'citable':
            self.citable = Citable(pathDF)
            self.dataframe = self.citable.digitalresource()

        if not pathType == 'citable':
            if self.datatype == 'pickle':
                self.dataframe = pd.read_pickle(pathDF)
            elif self.datatype == 'excel':
                self.dataframe = pd.read_excel(pathDF)
            elif self.datatype == 'csv':
                self.dataframe = pd.read_csv(pathDF, error_bad_lines=False)
            elif self.datatype == 'json':
                self.dataframe = pd.read_json(pathDF)
            else:
                raise ValueError(
                    'Please provide data in pickle,excel,csv or json format'
                    )

        self.dataindex = dataIndex

        self.colValueDictTrigger = []
        if self.dataindex == 'single':
            for col in self.dataframe.columns:
                if not self.dataframe[col].dtype.name in ['int', 'int64', 'float64']:
                    try:
                        length = len(self.dataframe[col].unique())
                        if length < maxValues:
                            self.colValueDictTrigger.append(col)
                    except TypeError:
                        print('Encountered list value in cell, skipping...')
                        pass
        elif self.dataindex == 'multi':
            for level in self.dataframe.index.names:
                if not self.dataframe.index.get_level_values(level).dtype.name in ['int', 'int64', 'float64']:
                    if len(self.dataframe.index.get_level_values(level).unique()) < maxValues:
                        self.colValueDictTrigger.append(level)

        self.extData = ''
        self.result = ''

        self.levelValues = {}

        self.column = colname

        if pathMeta:
            self.metaData = pd.read_excel(pathMeta)
        else:
            self.metaData = ''

    def resetColWidth(self):
        """ Reset pandas display option for max_colwidth"""
        pd.reset_option('display.max_colwidth')
        return

    def resetSearch(self):
        """Reset self.result to full dataframe."""
        self.result = ''
        return

    def search(self, value):
        """
        Search string in colname column.
        """
        if type(self.result) == str:
            self.result = self.dataframe[self.dataframe[self.column].str.contains(value)]
        elif isinstance(self.result, pd.DataFrame):
            self.result = self.result[self.result[self.column].str.contains(value)]
        return self

    def logicReduce(self, logicList):
        """
        Constructs complex searches for list of (part,value) tuples, connected
        via AND (&),OR (|),NOT (~&) logic. logicList has the format
        '[(part1,value1),&,(part2,value2),|,(part3,value3)]'.
        Evaluation is in order of apperance, e.g. for the above logicList,
        a boolean list is constructed for each (part,value) tuple.
        Then the first two tuples are compared with & yielding a temporary
        result temp1, which is compared with | with the last tuple.
        This creates a resulting boolean list res1,
        which is used to reduce the dataframe.
        """
        self.tempRes = []
        for part in logicList:
            if type(part) == tuple:
                res = self.boolList(part[0], part[1])
                self.tempRes.append(res)
            elif type(part) == str and part in ['&', '|', '~&']:
                self.tempRes.append(part)

        self.res = self.tempRes.pop(0)

        while self.tempRes:
            op = self.tempRes.pop(0)
            res2 = self.tempRes.pop(0)
            if op == '&':
                self.res = self.res & res2
            elif op == '|':
                self.res = self.res | res2
            elif op == '~&':
                self.res = self.res & np.invert(res2)
            else:
                pass

        if type(self.res) != str:
            self.result = self.dataframe[self.res]
        return self

    def boolList(self, level, value):
        """ Returns boolean list for dataframe[part]==value."""
        if self.dataindex == 'multi' and level != self.column:
            res = self.dataframe.index.get_level_values(level) == self._fuzzySearch(level, value)
        elif self.dataindex == 'single' and level != self.column:
            res = self.dataframe[level] == self._fuzzySearch(level, value)
        elif level == self.column:
            res = self.dataframe[level].str.contains(value) == True
        return res

    def _fuzzySearch(self, level, value):
        """Allow fuzzy search."""
        if level in self.colValueDictTrigger and level != self.column:
            if level not in self.levelValues.keys():
                if self.dataindex == 'multi':
                    self.levelValues[level] = self.dataframe.index.get_level_values(level).unique()
                elif self.dataindex == 'single':
                    self.levelValues[level] = self.dataframe[level].unique()
                else:
                    raise ValueError(
                        'DataIndex not "single" or "multi".'
                        )
            else:
                pass

            if value not in self.levelValues[level]:
                try:
                    closestMatch = difflib.get_close_matches(value, self.levelValues[level], 1)
                    if closestMatch:
                        searchValue = closestMatch[0]
                    else:
                        raise ValueError(
                            'Could not find matching expression to search.'
                            )
                except TypeError:
                    searchValue = value
            else:
                searchValue = value
        else:
            searchValue = value
        return searchValue

    def reduce(self, level, value):
        """ Return reduced dataframe for search tuple (level/column,value):
            a) as cross-section for multi-index dataframe:
               df.xs(value,level=level)
            b) as sub-dataframe for single-index dataframe:
               df[df.column == value]

            Result is in self.result, to be able to chain reductions.
            To view result use self.results()
        """

        searchValue = self._fuzzySearch(level, value)
        self._searchString(level, searchValue)
        return self

    def _searchString(self, level, value):
        """Helper function for reducing dataframes"""
        if self.dataindex == 'multi':
            if type(self.result) == str:
                self.result = self.dataframe.xs(self._assertDataType(level, value, self.dataframe), level=level, drop_level=False)
            else:
                self.result = self.result.xs(self._assertDataType(level, value, self.result), level=level, drop_level=False)
        elif self.dataindex == 'single':
            if type(self.result) == str:
                self.result = self.dataframe[self.dataframe[level] == self._assertDataType(level, value, self.dataframe)]
            else:
                self.result = self.result[self.result[level] == self._assertDataType(level, value, self.result)]

    def _assertDataType(self, level, value, dataframe):
        """Helper function to assert correct datatype for value at level"""
        intTypes = ['int8', 'int16', 'int32', 'int64', 'float', 'float64']
        valueType = type(value)
        if self.dataindex == 'multi':
            levelType = dataframe.index.get_level_values(level=level).dtype.name
        elif self.dataindex == 'single':
            levelType = self.dataframe[level].dtype.name
        if levelType == 'object' and valueType == str:
            return value
        elif levelType == 'object' and valueType == int:
            try:
                return str(value)
            except ValueError as err:
                raise('Can not cast {0} to type {1}'.format(value, levelType))
        elif levelType in intTypes and valueType == int:
            return value
        elif levelType in intTypes and valueType == float:
            return value
        elif levelType in intTypes and valueType == str:
            try:
                return int(value)
            except ValueError:
                raise('Can not cast {0} to type {1}'.format(value, levelType))
        else:
            raise ValueError('Can not cast {0} to type {1}'.format(value, levelType))

    def results(self):
        """Returns the search result as a single-index dataframe."""
        if self.dataindex == 'multi':
            self.indexLevels = list(self.result.index.names)
            formatedResult = self.result.reset_index(level=self.indexLevels)
        elif self.dataindex == 'single':
            formatedResult = self.result
        pd.set_option('display.max_colwidth', -1)
        return formatedResult

    def extResults(self, level):
        """
        Returns the search result as a single-index dataframe, extend with
        metadata on the desired level. The metadata is calculated for all other
        levels of multi-indexed dataframe.
        """
        if not self.dataindex == 'multi':
            print('Statistics for multiindexed dataframes only.')
            return
        self.extData = pd.merge(left=self.results(), right=self._metaData(level), on=level)
        cols = [x for x in self.extData.columns if x != self.column]
        cols = cols + [self.column]
        self.extData = self.extData[cols]
        pd.set_option('display.max_colwidth', -1)
        return self.extData

    def _countWords(self, level, value):
        """Helper function to count words on a given level."""
        text = ' '.join(self.dataframe.xs(value, level=level).text.tolist())
        numWords = len(re.findall('\w+', text))
        return numWords

    def _getStatistics(self, level):
        """Helper function to return statistics on given level."""
        retDict = {}
        subLevels = self.indexLevels.copy()
        try:
            subLevels.remove(level)
        except ValueError:
            print('{0} not in index'.format(level))

        for part in self.dataframe.index.get_level_values(level).unique():
            partDict = {}
            for subLevel in subLevels:
                num = len(self.dataframe.xs(part, level=level).index.get_level_values(subLevel).unique())
                if num > 1:
                    partDict['Num_' + subLevel] = num
            partDict['words'] = self._countWords(level, part)
            retDict[part] = partDict
        return retDict

    def _metaData(self, level):
        """Helper function to generate dataframe from statistics."""
        resDict = self._getStatistics(level)
        self.metaDataFrame = pd.DataFrame(resDict).fillna('1').astype('int').transpose().reset_index().rename({'index': level}, axis=1)
        return self.metaDataFrame



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()

setup(
    # Application name:
    name="corpussearch",

    # Version number (initial):
    version="0.0.10",

    # Application author details:
    author="Malte Vogl",
    author_email="mvogl@mpiwg-berlin.mpg.de",

    # Packages
    packages=find_packages(exclude=('tests')),

    # Include additional files into the package
    include_package_data=True,

    url='https://github.com/TOPOI-DH/corpussearch/',

    # Details

    license='GPLv3',
    description="Tools for loading and analyzing large text corpora.",

    long_description=long_description,

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Indexing',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],

    project_urls={
        'Home': 'https://github.com/TOPOI-DH/corpussearch/',
        'Tracker': 'https://github.com/TOPOI-DH/corpussearch/issues',
        'Download': 'https://github.com/TOPOI-DH/corpussearch/archive/0.0.10.tar.gz',
    },

    python_requires='>=3',

    # Dependent packages (distributions)
    install_requires=[
        "pandas",
        "numpy",
        "citableclass",
        "ipywidgets",
        "Ipython",
        "gensim",
        "cltk"
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
)
